# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Objects and methods for dealing with PyTorch data."""

from pathlib import Path
from typing import Dict, Sequence

import torch
import torch.utils.data
from Bio import File, SeqIO

from gcgc.alphabet import ExtendedIUPACDNAEncoding
from gcgc.alphabet.base import EncodingAlphabet
from gcgc.ml.pytorch_utils.parser import TorchSequenceParser
from gcgc.parser.gcgc_record import GCGCRecord


class _SequenceIndexer(object):
    """A helper object that is used to index multiple files at ones."""

    def __init__(self):
        """Initialize the _SequenceIndexer object."""
        self._record_index = {}
        self._counter = -1

    def __call__(self, sid):
        """Return the record index if it's known, otherwise generate a new one."""
        try:
            return self._record_index[sid]
        except KeyError:
            self._counter = self._counter + 1
            self._record_index[sid] = self._counter
            return self._record_index[sid]


class GenomicDataset(torch.utils.data.Dataset):
    """GenomicDataset can be used to load sequence information into a format aminable to PyTorch."""

    def __init__(
        self, file_index: Dict[Path, File._IndexedSeqFileDict], parser: TorchSequenceParser
    ):
        """Initialize the GenomicDataset object."""

        self._file_index = file_index
        self._parser = parser

        super().__init__()

    @classmethod
    def from_path(
        cls,
        path: Path,
        parser: TorchSequenceParser,
        file_format: str = "fasta",
        alphabet: EncodingAlphabet = ExtendedIUPACDNAEncoding(),
    ) -> "GenomicDataset":
        """Init from a single file. This is a convience method that delegates to from_paths."""

        return cls.from_paths([path], parser, file_format, alphabet)

    @classmethod
    def from_paths(
        cls,
        path_sequence: Sequence[Path],
        parser: TorchSequenceParser,
        file_format="fasta",
        alphabet: EncodingAlphabet = ExtendedIUPACDNAEncoding(),
    ) -> "GenomicDataset":
        """Initialize the GenomicDataset from a pathlib.Path sequence."""

        file_index = {}
        si = _SequenceIndexer()

        for f in sorted(path_sequence):
            file_index[f] = SeqIO.index(str(f), file_format, key_function=si, alphabet=alphabet)

        return cls(file_index, parser)

    def __len__(self) -> int:
        """Return the length of the dataset."""

        return sum(len(v) for v in self._file_index.values())

    def __getitem__(self, i: int):
        """Get the record from the index."""

        for k, v in self._file_index.items():
            try:
                r = GCGCRecord(path=k, seq_record=v[i])
                return self._parser.parse_record(r)
            except KeyError:
                pass

        else:
            raise RuntimeError(f"Exausted file index while looking for {i}.")



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

import unittest

from Bio import AlignIO
from Bio.Alphabet import IUPAC

from gcgc.encoded_seq import EncodedSeq
from gcgc.tests.fixtures import PF01152_PATH


class TestAlignment(unittest.TestCase):
    def test_read_alignment_file(self):

        protein_alphabet = IUPAC.ExtendedIUPACProtein()

        gaps = {".", "-"}

        # Times two because add_lower_case_for_inserts=True, 3 because start, end, padding.
        expected_dim = 2 * (len(protein_alphabet.letters)) + len(gaps) + 3

        with open(PF01152_PATH) as f:

            alignment = AlignIO.read(f, "stockholm", alphabet=protein_alphabet)
            for f in alignment:

                seq = EncodedSeq.from_seq(
                    f.seq, gap_characters=gaps, add_lower_case_for_inserts=True
                )

                one_hot_dim = len(seq.one_hot_encoded[0])
                self.assertEqual(one_hot_dim, expected_dim)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""A set of Datasets which are comprised of DataFiles."""

import asyncio
import logging
from pathlib import Path
from typing import List, Optional, Sequence

import aiohttp

logger = logging.getLogger(__name__)


class DataFile:
    """An external file download and place at the local path."""

    def __init__(self, url: Optional[str], local_path: Path) -> None:
        """Init the DataFile object."""

        self.url = url
        self.local_path = local_path

    async def download(self, session):
        """Given a session, query the url."""

        async with session.get(self.url) as response:
            with open(self.local_path, "wb") as f_handle:
                while True:
                    chunk = await response.content.read(1024)
                    if not chunk:
                        break
                    f_handle.write(chunk)

            logger.info(f"Downloaded {self.url} to {self.local_path}.")
            return await response.release()


DataFiles = List[DataFile]


class UniprotDataset(object):
    """A base class that loads data from Uniprot."""

    def __init__(self, data_files: DataFiles):
        """Initialize the UniprotDataset object."""

        self.data_files = data_files

    @staticmethod
    def _url(query_str, file_format):
        return f"https://www.uniprot.org/uniprot/?{query_str}&format={file_format}"

    def download_files(self):
        """Download the files."""
        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.wait([self._download()]))
        loop.close()

    async def _download(self):
        async with aiohttp.ClientSession() as session:
            downloaded_files = [f.download(session) for f in self.data_files]
            await asyncio.gather(*downloaded_files)


class TaxonDataset(UniprotDataset):
    """Download a set of taxons IDs from uniprot."""

    def __init__(
        self, taxon_ids: Sequence[int], data_directory: Path = Path("."), file_format: str = "fasta"
    ):
        """Initialize the TaxonDataset object."""

        self.taxon_ids = taxon_ids
        files_to_download: DataFiles = []

        for taxon_id in self.taxon_ids:

            url = self._url(f"query=reviewed:yes+AND+organism:{taxon_id}", file_format=file_format)
            local_path = data_directory / str(f"{taxon_id}.{file_format}")

            files_to_download.append(DataFile(url, local_path))

        super().__init__(files_to_download)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Holds the base EncodingAlphabet."""

from abc import ABC
from typing import Optional, Sequence, Set

from gcgc.exceptions import GCGCAlphabetLetterEncodingException


class EncodingAlphabet(ABC):
    """The Encoding Alphabet is meant to be a baseclass for other alphabets."""

    START: str = ">"
    END: str = "<"
    PADDING: str = "|"

    # Convince linting that EncodingAlphabet will have a letters attribute.
    letters: str

    def __init__(
        self, gap_characters: Optional[Set[str]] = None, add_lower_case_for_inserts: bool = False
    ):
        """Create the EncodingAlphabet object."""

        self._gap_characters = gap_characters or set([])
        self._add_lower_case_for_inserts = add_lower_case_for_inserts

        self.letters_and_tokens = (
            self.letters + self.START + self.END + self.PADDING + "".join(self._gap_characters)
        )

        if self._add_lower_case_for_inserts:
            self.letters_and_tokens = self.letters_and_tokens + self.letters.lower()

        self.encoding_index = {letter: idx for idx, letter in enumerate(self.letters_and_tokens)}
        self.decoding_index = {idx: letter for letter, idx in self.encoding_index.items()}

    def __len__(self) -> int:
        """Get the lenght of the Alphabet."""
        return len(self.letters_and_tokens)

    def encode_token(self, token: str) -> int:
        """Given a particular token, return the integer representation."""

        return self.encoding_index[token]

    def decode_token(self, int_token: int) -> str:
        """Decode a token. This is the inverse of encode_token."""

        return self.decoding_index[int_token]

    def integer_encode(self, seq: str) -> Sequence[int]:
        """Integer encode the sequence."""

        try:
            encoded = []
            for s in seq:
                encoded.append(self.encoding_index[s])
            return encoded

        except KeyError:
            raise GCGCAlphabetLetterEncodingException(f"{s} not in {self.encoding_index}")

    def integer_decode(self, int_seq: Sequence[int]) -> str:
        """Given a sequence of integers, convert it to a string."""

        return "".join(self.decode_token(s) for s in int_seq)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

import unittest

from Bio.Alphabet import IUPAC

from gcgc.alphabet import IUPACUnambiguousDNAEncoding
from gcgc.alphabet.utils import biopython_alphabet_to_gcgc_alphabet
from gcgc.exceptions import GCGCAlphabetLetterEncodingException


class TestAlphabet(unittest.TestCase):
    def test_biopython_alphabet_to_gcgc_alphabet(self):
        dna = IUPAC.IUPACUnambiguousDNA()

        klass = biopython_alphabet_to_gcgc_alphabet(dna)
        self.assertIsInstance(klass, IUPACUnambiguousDNAEncoding)

    def test_len(self):
        dna = IUPACUnambiguousDNAEncoding()
        self.assertEqual(len(dna), len(dna.letters_and_tokens))

    def test_decoding_index(self):
        dna = IUPACUnambiguousDNAEncoding()

        self.assertEqual(dna.decode_token(0), dna.decoding_index[0])

    def test_encoding_index(self):
        dna = IUPACUnambiguousDNAEncoding()

        self.assertEqual(dna.encode_token("A"), dna.encoding_index["A"])

    def test_raise_integer_encoding_exception(self):

        dna = IUPACUnambiguousDNAEncoding()

        with self.assertRaises(GCGCAlphabetLetterEncodingException):
            dna.integer_encode("Z")

    def test_integer_decode(self):

        dna = IUPACUnambiguousDNAEncoding()

        code = [0, 1, 2, 3]
        expected = "".join(dna.decode_token(s) for s in code)
        actual = dna.integer_decode(code)

        self.assertEqual(expected, actual)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Utils for working with alphabets."""

from gcgc import alphabet
from gcgc.exceptions import GCGCAlphabetException


def biopython_alphabet_to_gcgc_alphabet(biopython_alphabet_instance, *args, **kwargs):
    """Convert the BioPython Alphabet to the associated GCGC Alphabet."""

    # Might also try just creating the type on the fly.

    for gcgc_alphabet in alphabet.__all__:

        klass = getattr(alphabet, gcgc_alphabet)
        instance = klass(*args, **kwargs)

        if isinstance(instance, type(biopython_alphabet_instance)):
            return instance

    msg = f"No instance of {type(biopython_alphabet_instance)} among {alphabet.__all__}."
    raise GCGCAlphabetException(msg)
